local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTooltip", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("tomQuest2")
local LibQTip = LibStub("LibQTip-1.0")
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("tomQuest2", {type = "data source", icon = "Interface\\AddOns\\tomQuest2\\media\\icon"})
local defaults = {
	profile = {
		hideCompletedObjective = 3,
		headerFontSize = 12,
		normalFontSize = 10,
		maxwidth = 300,
		showCurrentQuest = true,
		showCompletedQuest = true,
		showMaxQuest = true,
		questscellpadding = 5,
		objectivescellpadding = 10,
		chatlink = true,
		questsobjectives = true,
		clicktooltip = true,
		sharequest = true,
		trackquest = true,
		abandonquest = true,
		questlog = true,
		overtooltip = true,
	}
}

local prdb, chdb
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTooltip", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	chdb.zoneDisplay = chdb.zoneDisplay or {}
end

module.modName = L["Broker Quests Tooltip"]

local options
function module:getOptions()
	if not options then
		options = {		
			showNumCompletedQuests = {
				name = L["Show number of completed quest"],
				type = "toggle",
				get = function(info)
					return prdb.showNumCompletedQuests
				end,
				set = function(info, value)
					prdb.showNumCompletedQuests = value
					module:updateLDBText()
				end,
				width = "full",
				order = 11,
			},
			showNumQuests = {
				name = L["Show number of quests"],
				type = "toggle",
				get = function(info)
					return prdb.showNumQuests
				end,
				set = function(info, value)
					prdb.showNumQuests = value
					module:updateLDBText()
				end,
				width = "full",
				order = 12,
			},
			showMaxQuest = {
				name = L["Show maximum quests number"],
				type = "toggle",
				get = function(info)
					return prdb.showMaxQuest
				end,
				set = function(info, value)
					prdb.showMaxQuest = value
					module:updateLDBText()
				end,
				width = "full",
				order = 13,
			},
			tooltipheader = {
				type = "header",
				name = L["Tooltip Settings"],
				order = 14,
			},
			hideCompletedObjective = {
				name = L["Hide completed quests Objectives"],
				type = "select",
				values = {[1] = L["When you've completed the objective"], [2] = L["When everyone in the party have completed the objective"], [3] = L["Never"]},
				get = function(info)
					return prdb.hideCompletedObjective
				end,
				set = function(info, value)
					prdb.hideCompletedObjective = value
				end,
				width = "full",
				order = 15,
			},
			headerFontSize = {
				name = L["Quests font size"],
				type="range",
				min = 6,
				max = 20,
				step = 1,
				get = function(info)
					return prdb.headerFontSize
				end,
				set = function(info, value)
					prdb.headerFontSize = value
				end,
				width = "full",
				order=16,
			},
			normalFontSize = { 
				name = L["Quests Objectives font size"],
				type="range",
				min = 6,
				max = 20,
				step = 1,
				get = function(info)
					return prdb.normalFontSize
				end,
				set = function(info, value)
					prdb.normalFontSize = value
				end,
				width = "full",
				order=17,
			},
			maxWidth = {
				name = L["Quests Tooltip column max width"],
				type="range",
				min = 50,
				max = 600,
				step = 5,
				get = function(info)
					return prdb.maxwidth
				end,
				set = function(info, value)
					prdb.maxwidth = value
				end,
				width = "full",
				order=18,
			},
			questscellpadding = {
				name = L["Quests cell padding"],
				type="range",
				min = 0,
				max = 50,
				step = 1,
				get = function(info)
					return prdb.questscellpadding
				end,
				set = function(info, value)
					prdb.questscellpadding = value
				end,
				width = "full",
				order=19,
			},
			objectivescellpadding = {
				name = L["Quests objectives cell padding"],
				type="range",
				min = 0,
				max = 50,
				step = 1,
				get = function(info)
					return prdb.objectivescellpadding
				end,
				set = function(info, value)
					prdb.objectivescellpadding = value
				end,
				width = "full",
				order=20,
			},
			tooltipinteraction = {
				type = "header",
				name = L["Tooltip interaction"],
				order = 21,
			},
			chatlink  = {
				name = L["Toggle chatlink when shift - right clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.chatlink
				end,
				set = function(info, value)
					prdb.chatlink = value
				end,
				width = "full",
				order = 22,
			},
			questsobjectives = {
				name = L["Toggle insert of quest objectives in chat when ctrl - right clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.questsobjectives
				end,
				set = function(info, value)
					prdb.questsobjectives = value
				end,
				width = "full",
				order = 23,
			},
			clicktooltip = {
				name = L["Toggle display of quest tooltip when right clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.clicktooltip
				end,
				set = function(info, value)
					prdb.clicktooltip = value
				end,
				width = "full",
				order = 24,
			},
			sharequest = {
				name = L["Toggle quest sharing when ctrl - left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.sharequest
				end,
				set = function(info, value)
					prdb.sharequest = value
				end,
				width = "full",
				order = 25,
			},
			trackquest = {
				name = L["Toggle quest tracking when shift - left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.trackquest
				end,
				set = function(info, value)
					prdb.trackquest = value
				end,
				width = "full",
				order = 26,
			},
			abandonquest = {
				name = L["Toggle quest abandon when alt - left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.abandonquest
				end,
				set = function(info, value)
					prdb.abandonquest = value
				end,
				width = "full",
				order = 27,
			},
			questlog = {
				name = L["Toggle questlog display when left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.questlog
				end,
				set = function(info, value)
					prdb.questlog = value
				end,
				width = "full",
				order = 28,
			},
			overtooltip = {
				name = L["Toggle quest tooltip when mouse over a quest"],
				type = "toggle",
				get = function(info)
					return prdb.overtooltip
				end,
				set = function(info, value)
					prdb.overtooltip = value
				end,
				width = "full",
				order = 29,
			},
		}
	end
	return options
end

local tip, tipAnchor
local tooltipIsShown = false
local maxQuests = MAX_QUESTS
local function hideTooltip()
	if MouseIsOver(tip) then return end
	tooltipIsShown = false
	tip:SetScript("OnLeave", nil)
	LibQTip:Release(tip)
	tip = nil
end

local timer = CreateFrame("Frame")

function module:OnEnable()
	core.RegisterCallback(self, "updateQuestsTooltip")
	core.RegisterCallback(self, "updateLDBText")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
end

function module:updateQuestsTooltip(event)
	if tooltipIsShown then
		tip:SetScript("OnLeave", nil)
		LibQTip:Release(tip)
		tip = nil
		module:updateTooltip()
	end
end

function module:updateLDBText(event)
	core.numQuests = core.numQuests or 0
	core.numCompletedQuests = core.numCompletedQuests or 0
	if prdb.showNumCompletedQuests then
		if prdb.showNumQuests then
			if prdb.showMaxQuest then
				if core.numQuests > 0 then
					dataobj.text = string.format("|cff%s%s/%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests, maxQuests)
				else
					dataobj.text = string.format("0/0/%s", maxQuests)
				end
			else
				if core.numQuests > 0 then
					dataobj.text = string.format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.numQuests), core.numCompletedQuests, core.numQuests)
				else
					dataobj.text = "0/0"
				end
			end
		elseif prdb.showMaxQuest then
			dataobj.text = string.format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numCompletedQuests/core.maxQuests), core.numCompletedQuests, maxQuests)
		else
			dataobj.text = core.numCompletedQuests
		end
	elseif prdb.showNumQuests then
		if prdb.showMaxQuest then
			dataobj.text = string.format("|cff%s%s/%s|r", core:getColorFromCompletionHEX(core.numQuests/maxQuests), core.numQuests, maxQuests)
		else
			dataobj.text = core.numQuests
		end	
	elseif prdb.showMaxQuest then
		dataobj.text = maxQuests
	else
		dataobj.text = ""
	end
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local plus = [[|TInterface\Buttons\UI-PlusButton-Up:24|t]]
local minus = [[|TInterface\Buttons\UI-MinusButton-Up:24|t]]
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, btype, bvalue, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame, button)
		if btype == "zone" then
			chdb.zoneDisplay[bvalue] = chdb.zoneDisplay[bvalue] == 1 and 0 or 1
			tip:SetScript("OnLeave", nil)
			LibQTip:Release(tip)
			tip = nil
			module:updateTooltip()
		elseif btype == "quest" then
			if button == "RightButton" then
				if ChatFrameEditBox:IsVisible() then
					if IsShiftKeyDown() then
						if core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.chatlink then
							ChatEdit_InsertLink(GetQuestLink(core.quests[bvalue].id))
						end
					elseif IsControlKeyDown() then
						if core.quests[bvalue].numobjectives > 0 and core.quests[bvalue].party[core.playerName] and prdb.questsobjectives then
							for objective in pairs(core.questsObjectives[bvalue]) do
								ChatFrameEditBox:Insert(format("{%s %s/%s} ", core.questsObjectives[bvalue][objective].desc, core.questsObjectives[bvalue][objective].party[core.playerName].got, core.questsObjectives[bvalue][objective].need))
							end
						end
					end
				elseif core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.clicktooltip then
					ItemRefTooltip:ClearLines()
					ShowUIPanel(ItemRefTooltip)
					if not ItemRefTooltip:IsVisible() then
						ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
					end
					local link = GetQuestLink(core.quests[bvalue].id)
					if link then
						ItemRefTooltip:SetHyperlink(link)
					end
				end
			elseif button == "LeftButton" then
				if IsControlKeyDown() then
					if core.numPartyMembers > 0 and core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.sharequest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bvalue].id)
						if GetQuestLogPushable() then
							QuestLogPushQuest()
						end
						SelectQuestLogEntry(wasSelected)
					end
				elseif IsShiftKeyDown() then
					if prdb.trackquest then
						core.cbh:Fire("toggleTrackedQuest", bvalue, true)
					end
				elseif IsAltKeyDown() then
					if core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.abandonquest then
						local wasSelected = GetQuestLogSelection()
						SelectQuestLogEntry(core.quests[bvalue].id)
						SetAbandonQuest()
						AbandonQuest()
						SelectQuestLogEntry(wasSelected)
					end
				elseif core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.questlog then
					if QuestLogFrame:IsVisible() then
						if GetQuestLogSelection() == core.quests[bvalue].id then
							return HideUIPanel(QuestLogFrame)
						end
					else
						ShowUIPanel(QuestLogFrame)
					end
					local numEntries = GetNumQuestLogEntries()
					if (numEntries > QUESTS_DISPLAYED) then
						if (core.quests[bvalue].id <= numEntries - QUESTS_DISPLAYED) then
							FauxScrollFrame_SetOffset(QuestLogListScrollFrame, core.quests[bvalue].id - 1)
							QuestLogListScrollFrameScrollBar:SetValue((core.quests[bvalue].id - 1) * QUESTLOG_QUEST_HEIGHT)
						else
							FauxScrollFrame_SetOffset(QuestLogListScrollFrame, numEntries - QUESTS_DISPLAYED)
							QuestLogListScrollFrameScrollBar:SetValue((numEntries - QUESTS_DISPLAYED) * QUESTLOG_QUEST_HEIGHT)
						end
					end
					SelectQuestLogEntry(core.quests[bvalue].id)
					QuestLog_SetSelection(core.quests[bvalue].id)
					_G["QuestLogTitle"..(core.quests[bvalue].id - FauxScrollFrame_GetOffset(QuestLogListScrollFrame))]:Click()
				end
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if prdb.overtooltip and btype == "quest" and core.quests[bvalue].id then
			local link = GetQuestLink(core.quests[bvalue].id)
			if link then
				GameTooltip:SetOwner(tooltip, anchor)
				GameTooltip:SetHyperlink(link)
			end
		end
	end)
	self:SetScript("OnLeave", function()
		if prdb.overtooltip and btype == "quest" and core.quests[bvalue].party[core.playerName] then
			GameTooltip:Hide()
		end
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, arg)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateTooltip()
	tip = LibQTip:Acquire("tomQuest2ToolTip", 2, "LEFT", "RIGHT")
	tooltipIsShown = true
	tip:Clear()
	tip:EnableMouse(true)
	tip:SetScript("OnLeave", hideTooltip)
	local headerFont = CreateFont("headerFont")
	headerFont:SetFont(GameTooltipText:GetFont(), prdb.headerFontSize)
	tip:SetHeaderFont(headerFont)
	local smallFont = CreateFont("smallFont")
	smallFont:SetFont(GameTooltipText:GetFont(), prdb.normalFontSize)
	tip:SetFont(smallFont)
	local tempLine
	tempLine = tip:AddLine()
	tip:SetCell(tempLine, 1, "tomQuest2", headerFont, "CENTER", 2, cProvider, "other", nil, prdb.maxwidth)
	tip:AddLine(" ")
	for zone in pairs(core.questsZones) do
		chdb.zoneDisplay[zone] = chdb.zoneDisplay[zone] or 1
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, string.format("%s |cff%s%s %s|r", chdb.zoneDisplay[zone] == 0 and plus or minus, core.questsZones[zone].color, zone, core.questsZones[zone].level ), headerFont, "LEFT", 2, cProvider, "zone", zone, prdb.maxwidth, 0)
		if chdb.zoneDisplay[zone] == 1 then
			for uid in pairs(core.questsByZone[zone].quests) do
				tempLine = tip:AddLine()
				tip:SetCell(tempLine, 1, format("|cff%s%s %s|r", core.quests[uid].color, core.quests[uid].taggedname, core.quests[uid].playeronquest > 1 and format("<%d>", core.quests[uid].playeronquest) or ""), headerFont, "LEFT", 1, cProvider, "quest", uid, prdb.maxwidth, prdb.questscellpadding, tipAnchor)
				tip:SetCell(tempLine, 2, core.quests[uid].pqstatus, headerFont, "RIGHT", 1, cProvider, "quest", uid, prdb.maxwidth, 0, tipAnchor)
				if core.questsObjectives[uid] then
					for objective in pairs(core.questsObjectives[uid]) do
						if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and not core.questsObjectives[uid][objective].partycomplete) or (prdb.hideCompletedObjective == 1 and not core.questsObjectives[uid][objective].playercomplete) then
							tempLine = tip:AddLine()
							tip:SetCell(tempLine, 1, format("|cff%s%s|r", core.questsObjectives[uid][objective].qocolor, core.questsObjectives[uid][objective].desc), smallFont, "LEFT", 1, cProvider, "quest", uid, prdb.maxwidth, prdb.objectivescellpadding, tipAnchor)
							tip:SetCell(tempLine, 2, core.questsObjectives[uid][objective].qostatus, smallFont, "RIGHT", 1, cProvider, "quest", uid, prdb.maxwidth, 0, tipAnchor)
						end
					end
				else
					tempLine = tip:AddLine()
					tip:SetCell(tempLine, 1, core.quests[uid].objective, smallFont, "LEFT", 2, cProvider,"quest", uid, prdb.maxwidth, prdb.objectivescellpadding, tipAnchor)
				end
			end
		end
	end
	tip:AddLine(" ")
	if prdb.questlog then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, L["|cffeda55fLeft click|r |cff19ff19Open quests log.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxwidth)
	end
	if prdb.trackquest then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, L["|cffeda55fShift + Left click|r |cff19ff19Add quest to tracker.|r"] , smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxwidth)
	end
	if prdb.sharequest then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, L["|cffeda55fCtrl + Left click|r |cff19ff19Share quest.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxwidth)
	end
	if prdb.abandonquest then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, L["|cffeda55fAlt+ Left click|r |cff19ff19Abandon quest.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxwidth)
	end
	if prdb.clicktooltip then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, L["|cffeda55fRight click|r |cff19ff19Show quest tooltip.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxwidth)
	end
	if prdb.chatlink then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, L["|cffeda55fShift + Right click|r |cff19ff19Send quest link to chat.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxwidth)
	end
	if prdb.questsobjectives then
		tempLine = tip:AddLine()
		tip:SetCell(tempLine, 1, L["|cffeda55fCtrl + Right click|r |cff19ff19Send quest progress to chat.|r"], smallFont, "LEFT", 2, cProvider, "other", nil, prdb.maxwidth)
	end
	tip:Show()
end

function dataobj.OnEnter(self)
	module:updateTooltip()
	tipAnchor = self:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	tip:SmartAnchorTo(self)
end

function dataobj.OnLeave(self)
	hideTooltip()
end

function dataobj.OnClick(self,button)
	if button == "RightButton" then
		return InterfaceOptionsFrame_OpenToCategory(LibStub("AceConfigDialog-3.0").BlizOptions["tomQuest2"].frame)
	else
		ToggleFrame(QuestLogFrame)
	end
end