local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("questsTracker", "AceEvent-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("tomQuest2")
local LibQTip = LibStub("LibQTip-1.0")
local defaults = {
	profile = {
		hideCompletedObjective = 3,
		onObjectiveProgress = true,
		onSubZone = true,
		point = "CENTER",
		x = 0,
		y = 0,
		showanchor = true,
		headerFontSize = 12,
		normalFontSize = 10,
		backDropColor = {0.09, 0.09, 0.09, 1},
		borderColor = {1, 1, 1, 1},
		scale = 1,
		maxwidth = 300,
		objectivescellpadding = 5,
		chatlink = true,
		questsobjectives = true,
		clicktooltip = true,
		sharequest = true,
		trackquest = true,
		abandonquest = true,
		questlog = true,
		overtooltip = true,
	},
}

local prdb, chdb
function module:OnInitialize()
	self.db = core.db:RegisterNamespace("questsTracker", defaults, "Default")
	prdb = self.db.profile
	chdb = self.db.char
	chdb.questsTracked = chdb.questsTracked or {}
	chdb.questsAutoTracked = chdb.questsAutoTracked or {}
end
local subZone = ""
module.modName = L["Quests Tracker"]

local numTrackedQuests = 0
local format = string.format
local tracker, trackerAnchor
local anchor = CreateFrame("Button", nil, UIParent)
anchor:SetHeight(24)
anchor:SetBackdrop({bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16, insets = {left = 5, right = 5, top = 5, bottom = 5}, tile = true, tileSize = 16})
anchor:SetBackdropColor(0.09, 0.09, 0.19, 0.5)
anchor:SetBackdropBorderColor(0.5, 0.5, 0.5, 1)
local text = anchor:CreateFontString(nil, nil, "GameFontNormalSmall")
text:SetPoint("CENTER")
text:SetText(L["Tracker"])
anchor:SetWidth(text:GetStringWidth() + 8)
anchor:SetMovable(true)
anchor:RegisterForDrag("LeftButton")
anchor:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
anchor:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	prdb.point, prdb.x, prdb.y = "BOTTOMLEFT", self:GetLeft(), self:GetBottom()
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	LibQTip:Release(tracker)
	module:updateQuestTracker()
end)


local options
function module:getOptions()
	if not options then
		options = {
			onObjectiveProgress = {
				name = L["Auto track quests on objective progress"],
				type = "toggle",
				get = function(info)
					return prdb.onObjectiveProgress
				end,
				set = function(info, value)
					prdb.onObjectiveProgress = value
				end,
				width = "full",
				order = 10,
			},
			onSubZone = {
				name = L["Auto track quests on sub zone change"],
				type = "toggle",
				get = function(info)
					return prdb.onSubZone
				end,
				set = function(info, value)
					prdb.onSubZone = value
				end,
				width = "full",
				order = 11,
			},
			showAnchor = {
				name = L["Show quest tracker anchor"],
				type = "toggle",
				get = function(info)
					return prdb.showanchor
				end,
				set = function(info, value)
					prdb.showanchor = value
					if prdb.showanchor then
						anchor:Show()
					else
						anchor:Hide()
					end
				end,
				width = "full",
				order = 12,
			},
			trackerheader = {
				type = "header",
				name = L["Tracker Settings"],
				order = 13,
			},
			hideCompletedObjective = {
				name = L["Hide completed quests Objectives"],
				type = "select",
				values = {[1] = L["When you've completed the objective"], [2] = L["When everyone in the party have completed the objective"], [3] = L["Never"]},
				get = function(info)
					return prdb.hideCompletedObjective
				end,
				set = function(info, value)
					prdb.hideCompletedObjective = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 14,
			},
			headerFontSize = {
				name = L["Quests font size"],
				type="range",
				min = 6,
				max = 20,
				step = 1,
				get = function(info)
					return prdb.headerFontSize
				end,
				set = function(info, value)
					prdb.headerFontSize = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order=15,
			},
			normalFontSize = { 
				name = L["Quests Objectives font size"],
				type="range",
				min = 6,
				max = 20,
				step = 1,
				get = function(info)
					return prdb.normalFontSize
				end,
				set = function(info, value)
					prdb.normalFontSize = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order=16,
			},
			maxWidth = {
				name = L["Quests Tracker column max width"],
				type="range",
				min = 50,
				max = 600,
				step = 5,
				get = function(info)
					return prdb.maxwidth
				end,
				set = function(info, value)
					prdb.maxwidth = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order=17,
			},
			objectivescellpadding = {
				name = L["Quests objectives cell padding"],
				type="range",
				min = 0,
				max = 50,
				step = 1,
				get = function(info)
					return prdb.objectivescellpadding
				end,
				set = function(info, value)
					prdb.objectivescellpadding = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order=18,
			},
			scale = {
				name = L["Quests Tracker scale"],
				type="range",
				min = 0,
				max = 1,
				step = 0.1,
				get = function(info)
					return prdb.scale
				end,
				set = function(info, value)
					prdb.scale = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order=19,
			},
			backDropColor = {
				name = L["Quests Tracker background color"],
				type = "color",
				hasAlpha = true,
				get = function(info)
					return unpack(prdb.backDropColor)
				end,
				set = function(info, r, g, b, a)
					prdb.backDropColor[1] = r
					prdb.backDropColor[2] = g
					prdb.backDropColor[3] = b
					prdb.backDropColor[4] = a
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 20,
			},
			borderColor = {
				name = L["Quests Tracker border color"],
				type = "color",
				hasAlpha = true,
				get = function(info)
					return unpack(prdb.borderColor)
				end,
				set = function(info, r, g, b, a)
					prdb.borderColor[1] = r
					prdb.borderColor[2] = g
					prdb.borderColor[3] = b
					prdb.borderColor[4] = a
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 21,
			},
			trackerinteraction = {
				type = "header",
				name = L["Tracker interaction"],
				order = 22,
			},
			chatlink  = {
				name = L["Toggle chatlink when shift - right clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.chatlink
				end,
				set = function(info, value)
					prdb.chatlink = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 23,
			},
			questsobjectives = {
				name = L["Toggle insert of quest objectives in chat when ctrl - right clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.questsobjectives
				end,
				set = function(info, value)
					prdb.questsobjectives = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 24,
			},
			clicktooltip = {
				name = L["Toggle display of quest tooltip when right clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.clicktooltip
				end,
				set = function(info, value)
					prdb.clicktooltip = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 25,
			},
			sharequest = {
				name = L["Toggle quest sharing when ctrl - left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.sharequest
				end,
				set = function(info, value)
					prdb.sharequest = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 26,
			},
			trackquest = {
				name = L["Toggle quest tracking when shift - left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.trackquest
				end,
				set = function(info, value)
					prdb.trackquest = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 27,
			},
			abandonquest = {
				name = L["Toggle quest abandon when alt - left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.abandonquest
				end,
				set = function(info, value)
					prdb.abandonquest = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 28,
			},
			questlog = {
				name = L["Toggle questlog display when left clicking on a quest"],
				type = "toggle",
				get = function(info)
					return prdb.questlog
				end,
				set = function(info, value)
					prdb.questlog = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 29,
			},
			overtooltip = {
				name = L["Toggle quest tooltip when mouse over a quest"],
				type = "toggle",
				get = function(info)
					return prdb.overtooltip
				end,
				set = function(info, value)
					prdb.overtooltip = value
					LibQTip:Release(tracker)
					module:updateQuestTracker()
				end,
				width = "full",
				order = 30,
			},
		}
	end
	return options
end

function module:OnEnable()
	anchor:SetPoint(prdb.point, prdb.x, prdb.y)
	if not prdb.showanchor then anchor:Hide() else anchor:Show() end
	trackerAnchor = anchor:GetCenter() > UIParent:GetWidth()*0.5 and "ANCHOR_LEFT" or "ANCHOR_RIGHT"
	self:RegisterEvent("MINIMAP_ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "zoneChanged")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "zoneChanged")
	self:SecureHook("AddQuestWatch")
	self:SecureHook("RemoveQuestWatch")
	self:SecureHook("QuestLogTitleButton_OnClick")
	self:SecureHook("QuestWatch_Update")
	core.RegisterCallback(self, "initTrackedQuest")
	core.RegisterCallback(self, "addPlayerQuest")
	core.RegisterCallback(self, "playerQuestComplete", "playerQuestChanged")
	core.RegisterCallback(self, "playerQuestFailed", "playerQuestChanged")
	core.RegisterCallback(self, "partyQuestComplete", "partyQuestChanged")
	core.RegisterCallback(self, "partyQuestFailed", "partyQuestChanged")
	core.RegisterCallback(self, "addPlayerQuestObjective")
	core.RegisterCallback(self, "addPartyQuestObjective")
	core.RegisterCallback(self, "playerQuestObjectiveProgress")
	core.RegisterCallback(self, "partyQuestObjectiveProgress")
	core.RegisterCallback(self, "playerQuestLost")
	core.RegisterCallback(self, "partyQuestLost")	
	core.RegisterCallback(self, "toggleTrackedQuest")
	core.RegisterCallback(self, "questObjectiveAdded")
	core.RegisterCallback(self, "questDescriptionAdded")
	core.RegisterCallback(self, "updateQuestsTracker")
	if numTrackedQuests ~= 0 then
		module:updateQuestTracker()
	end
end

function module:OnDisable()
	anchor:Hide()
	LibQTip:Release(tracker)
	tracker = nil
	core.UnregisterAllCallbacks(self)
end
--------------
--- Events ---
--------------
function module:zoneChanged()
	if UnitIsDead("player") then return end
	local trackerNeedUpdate = false
	subZone = GetSubZoneText() or ""
	for uid in pairs(chdb.questsAutoTracked) do
		chdb.questsAutoTracked[uid] = nil
		chdb.questsTracked[uid] = nil
		numTrackedQuests = numTrackedQuests - 1
		trackerNeedUpdate = true
	end
	if prdb.onSubZone and subZone ~= "" then
		for uid in pairs(core.quests) do
			if not chdb.questsTracked[uid] and ((core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true)) or (core.quests[uid].description and core.quests[uid].description:find(subZone, 1, true))) then
				chdb.questsAutoTracked[uid] = true
				chdb.questsTracked[uid] = true
				trackerNeedUpdate = true
				numTrackedQuests = numTrackedQuests + 1
			end
		end
	end
	if trackerNeedUpdate then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end
-------------
--- Hooks ---
-------------
function module:AddQuestWatch(id)
	local uid = core.questsIdToUid[id]
	if not chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = true
		chdb.questsAutoTracked[uid] = nil
		numTrackedQuests = numTrackedQuests + 1
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:RemoveQuestWatch(id)
	local uid = core.questsIdToUid[id]
	if uid and chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsAutoTracked[uid] = nil
		numTrackedQuests = numTrackedQuests - 1
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:QuestLogTitleButton_OnClick(button)
	if IsModifiedClick("QUESTWATCHTOGGLE") then
		local uid = core.questsIdToUid[this:GetID() + FauxScrollFrame_GetOffset(QuestLogListScrollFrame)]
		if not uid then return end
		if chdb.questsTracked[uid] then
			chdb.questsTracked[uid] = nil
			chdb.questsAutoTracked[uid] = nil
			numTrackedQuests = numTrackedQuests - 1
		else
			chdb.questsTracked[uid] = true
			chdb.questsAutoTracked[uid] = nil
			numTrackedQuests = numTrackedQuests + 1
		end
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:QuestWatch_Update()
	QuestWatchFrame:Hide()
end
-----------------
--- Callbacks ---
-----------------
function module:initTrackedQuest(event)
	for uid in pairs(chdb.questsTracked) do
		if not core.quests[uid] then
			chdb.questsTracked[uid] = nil
		else
			numTrackedQuests = numTrackedQuests + 1
		end
	end
	for uid in pairs(chdb.questsAutoTracked) do
		if not core.quests[uid] then
			chdb.questsAutoTracked[uid] = nil
		end
	end
	LibQTip:Release(tracker)
	module:updateQuestTracker()
end

function module:addPlayerQuest(event, first, commHandler, uid, title, shorttag, level, numobjectives, complete, zone)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and ((core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true)) or (core.quests[uid].description and core.quests[uid].description:find(subZone, 1, true))) then
			chdb.questsTracked[uid] = true
			chdb.questsAutoTracked[uid] = true
			numTrackedQuests = numTrackedQuests + 1
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	end
end

function module:playerQuestChanged(event, first, commHandler, uid)
	if chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:partyQuestChanged(event, sender, uid)
	if chdb.questsTracked[uid] then
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:addPlayerQuestObjective(event, first, commHandler, uid, objective, desc, qtype, got, need, ngot, nneed)
	if prdb.onObjectiveProgress and ngot ~= 0 then
		if not chdb.questsTracked[uid] then
			chdb.questsTracked[uid] = true
			numTrackedQuests = numTrackedQuests + 1
		elseif chdb.questsAutoTracked[uid] then
			chdb.questsAutoTracked[uid] = nil
		end
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:addPartyQuestObjective(event, uid)
	if prdb.onObjectiveProgress then
		if not chdb.questsTracked[uid] then
			chdb.questsTracked[uid] = true
			numTrackedQuests = numTrackedQuests + 1
		elseif chdb.questsAutoTracked[uid] then
			chdb.questsAutoTracked[uid] = nil
		end
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:playerQuestObjectiveProgress(event, first, commHandler, uid, objective, got, ngot)
	if prdb.onObjectiveProgress then
		if not chdb.questsTracked[uid] then
			chdb.questsTracked[uid] = true
			numTrackedQuests = numTrackedQuests + 1
		elseif chdb.questsAutoTracked[uid] then
			chdb.questsAutoTracked[uid] = nil
		end
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:partyQuestObjectiveProgress(event, sender, uid, objective, got, ngot)
	if prdb.onObjectiveProgress then
		if not chdb.questsTracked[uid] then
			chdb.questsTracked[uid] = true
			numTrackedQuests = numTrackedQuests + 1
		elseif chdb.questsAutoTracked[uid] then
			chdb.questsAutoTracked[uid] = nil
		end
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:playerQuestLost(event, first, commHandler, uid)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsAutoTracked[uid] = nil
		numTrackedQuests = numTrackedQuests - 1
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:partyQuestLost(event, uid)
	if not core.quests[uid] and chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsAutoTracked[uid] = nil
		numTrackedQuests = numTrackedQuests - 1
		LibQTip:Release(tracker)
		module:updateQuestTracker()
	end
end

function module:toggleTrackedQuest(event, uid)
	if chdb.questsTracked[uid] then
		chdb.questsTracked[uid] = nil
		chdb.questsAutoTracked[uid] = nil
		numTrackedQuests = numTrackedQuests - 1
	else
		chdb.questsTracked[uid] = true
		numTrackedQuests = numTrackedQuests + 1
	end
	LibQTip:Release(tracker)
	module:updateQuestTracker()
end

function module:questObjectiveAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].objective and core.quests[uid].objective:find(subZone, 1, true) then
			chdb.questsTracked[uid] = true
			chdb.questsAutoTracked[uid] = true
			numTrackedQuests = numTrackedQuests + 1
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	end
end

function module:questDescriptionAdded(event, uid)
	if prdb.onSubZone and subZone ~= "" then
		if not chdb.questsTracked[uid] and core.quests[uid].description	and core.quests[uid].description:find(subZone, 1, true) then
			chdb.questsTracked[uid] = true
			chdb.questsAutoTracked[uid] = true
			numTrackedQuests = numTrackedQuests + 1
			LibQTip:Release(tracker)
			module:updateQuestTracker()
		end
	end
end

function module:updateQuestsTracker(event)
	LibQTip:Release(tracker)
	module:updateQuestTracker()
end

----------------------------
--- Custom Cell Provider ---
----------------------------
local cProvider, cPrototype, basecCellPrototype = LibQTip:CreateCellProvider(LibQTip.LabelProvider)
function cPrototype:InitializeCell()
	self.fontString = self:CreateFontString()
	self.fontString:SetFontObject(GameTooltipText)
end

function cPrototype:SetupCell(tooltip, value, justification, font, bvalue, maxWidth, indent, anchor)
	self:EnableMouse(true)
	self:SetScript("OnMouseDown", function(frame,button)
		if button == "RightButton" then
			if ChatFrameEditBox:IsVisible() then
				if IsShiftKeyDown() then
					if core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.chatlink then
						ChatEdit_InsertLink(GetQuestLink(core.quests[bvalue].id))
					end
				elseif IsControlKeyDown() then
					if core.quests[bvalue].numobjectives > 0 and core.quests[bvalue].party[core.playerName] and prdb.questsobjectives then
						for objective in pairs(core.questsObjectives[bvalue]) do
							ChatFrameEditBox:Insert(format("{%s %s/%s} ", core.questsObjectives[bvalue][objective].desc, core.questsObjectives[bvalue][objective].party[core.playerName].got, core.questsObjectives[bvalue][objective].need))
						end
					end
				end
			elseif core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.clicktooltip then
				ItemRefTooltip:ClearLines()
				ShowUIPanel(ItemRefTooltip)
				if not ItemRefTooltip:IsVisible() then
					ItemRefTooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
				end
				local link = GetQuestLink(core.quests[bvalue].id)
				if link then
					ItemRefTooltip:SetHyperlink(link)
				end
			end
		elseif button == "LeftButton" then
			if IsControlKeyDown() then
				if core.numPartyMembers > 0 and core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.sharequest then
					local wasSelected = GetQuestLogSelection()
					SelectQuestLogEntry(core.quests[bvalue].id)
					if GetQuestLogPushable() then
						QuestLogPushQuest()
					end
					SelectQuestLogEntry(wasSelected)
				end
			elseif IsShiftKeyDown() then
				if prdb.trackquest then
					module:toggleTrackedQuest(nil, bvalue, true)
				end
			elseif IsAltKeyDown() then
				if core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.abandonquest then
					local wasSelected = GetQuestLogSelection()
					SelectQuestLogEntry(core.quests[bvalue].id)
					SetAbandonQuest()
					AbandonQuest()
					SelectQuestLogEntry(wasSelected)
				end
			elseif core.quests[bvalue].party[core.playerName] and core.quests[bvalue].id and prdb.questlog and prdb.questlog then
				if QuestLogFrame:IsVisible() then
					if GetQuestLogSelection() == core.quests[bvalue].id then
						return HideUIPanel(QuestLogFrame)
					end
				else
					ShowUIPanel(QuestLogFrame)
				end
				local numEntries = GetNumQuestLogEntries()
				if (numEntries > QUESTS_DISPLAYED) then
					if (core.quests[bvalue].id <= numEntries - QUESTS_DISPLAYED) then
						FauxScrollFrame_SetOffset(QuestLogListScrollFrame, core.quests[bvalue].id - 1)
						QuestLogListScrollFrameScrollBar:SetValue((core.quests[bvalue].id - 1) * QUESTLOG_QUEST_HEIGHT)
					else
						FauxScrollFrame_SetOffset(QuestLogListScrollFrame, numEntries - QUESTS_DISPLAYED)
						QuestLogListScrollFrameScrollBar:SetValue((numEntries - QUESTS_DISPLAYED) * QUESTLOG_QUEST_HEIGHT)
					end
				end
				SelectQuestLogEntry(core.quests[bvalue].id)
				QuestLog_SetSelection(core.quests[bvalue].id)
				_G["QuestLogTitle"..(core.quests[bvalue].id - FauxScrollFrame_GetOffset(QuestLogListScrollFrame))]:Click()
			end
		end
	end)
	self:SetScript("OnEnter", function()
		if prdb.overtooltip and core.quests[bvalue].id then
			local link = GetQuestLink(core.quests[bvalue].id)
			if link then
				GameTooltip:SetOwner(tooltip, anchor)
				GameTooltip:SetHyperlink(link)
			end
		end
	end)
	self:SetScript("OnLeave", function()
		if prdb.overtooltip and core.quests[bvalue].party[core.playerName] then
			GameTooltip:Hide()
		end
	end)

	indent = indent or 0
	local width, height = basecCellPrototype.SetupCell(self, tooltip, value, justification, font, arg)
	local fs = self.fontString
	fs:SetPoint("TOPLEFT", self, "TOPLEFT", indent, 0)
	fs:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", 0, 0)
	width = width + indent
	if width > maxWidth then
		fs:SetWidth(maxWidth)
		height = fs:GetHeight()
		width = maxWidth
	end
	fs:Show()
	return width, height
end

function cPrototype:ReleaseCell() 
	self:EnableMouse(false) 
	self:ClearAllPoints()
	self:SetScript("OnEnter", nil) 
	self:SetScript("OnLeave", nil) 
	self:SetScript("OnMouseDown", nil) 
end

function module:updateQuestTracker()
	if numTrackedQuests == 0 then return end
	tracker = LibQTip:Acquire("tomQuest2Tracker", 2, "LEFT", "RIGHT")
	tracker:EnableMouse(true)
	tracker:SetBackdropColor(prdb.backDropColor[1], prdb.backDropColor[2], prdb.backDropColor[3], prdb.backDropColor[4])
	tracker:SetBackdropBorderColor(prdb.borderColor[1], prdb.borderColor[2], prdb.borderColor[3], prdb.borderColor[4])
	tracker:SetScale(prdb.scale)
	tracker:SetFrameStrata("BACKGROUND")

	local headerFont = CreateFont("headerFont")
	headerFont:SetFont(GameTooltipText:GetFont(), prdb.headerFontSize)
	tracker:SetHeaderFont(headerFont)

	local smallFont = CreateFont("smallFont")
	smallFont:SetFont(GameTooltipText:GetFont(), prdb.normalFontSize)
	tracker:SetFont(smallFont)

	local tempLine
	for uid in pairs(chdb.questsTracked) do
		if uid and core.quests[uid] then
			tempLine = tracker:AddLine()
			tracker:SetCell(tempLine, 1, format("|cff%s%s %s|r", core.quests[uid].color, core.quests[uid].taggedname, core.quests[uid].playeronquest > 1 and format("<%d>", core.quests[uid].playeronquest) or ""), headerFont, "LEFT", 1, cProvider, uid, prdb.maxwidth, 0, trackerAnchor)
			tracker:SetCell(tempLine, 2, core.quests[uid].pqstatus, headerFont, "RIGHT", 1, cProvider, uid, prdb.maxwidth, 0, trackerAnchor)
			if core.questsObjectives[uid] then
				for objective in pairs(core.questsObjectives[uid]) do
					if prdb.hideCompletedObjective == 3 or (prdb.hideCompletedObjective == 2 and not core.questsObjectives[uid][objective].partycomplete) or (prdb.hideCompletedObjective == 1 and not core.questsObjectives[uid][objective].playercomplete) then
						tempLine = tracker:AddLine()
						tracker:SetCell(tempLine, 1, format("|cff%s%s|r", core.questsObjectives[uid][objective].qocolor, core.questsObjectives[uid][objective].desc), smallFont, "LEFT", 1, cProvider, uid, prdb.maxwidth, prdb.objectivescellpadding, trackerAnchor)
						tracker:SetCell(tempLine, 2, core.questsObjectives[uid][objective].qostatus, smallFont, "RIGHT", 1, cProvider, uid, prdb.maxwidth, 0, trackerAnchor)
					end
				end
			else
				tempLine = tracker:AddLine()
				tracker:SetCell(tempLine, 1, core.quests[uid].objective, smallFont, "LEFT", 2, cProvider, uid, prdb.maxwidth, 20, trackerAnchor)
			end
		end
	end
	tracker:SmartAnchorTo(anchor)
	tracker:Show()
end