local core = LibStub("AceAddon-3.0"):GetAddon("tomQuest2")
local module = core:NewModule("syncParty", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("tomQuest2")
local addonPrefix = "tomQ2"
local version = 3
local firstCompatibleVersion = 3
local lastIncompatibleVersion = 2
local syncParty = false

module.modName = L["Party Synchronization"]
local function delTable(t)
	for k, v in pairs(t) do
		if (type(v) == "table") then
			delTable(v);
		end
		t[k] = nil;
	end
	t = nil;
end

function module:OnEnable()
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterComm(addonPrefix, "commHandler")
 	core.RegisterCallback(self, "addPlayerQuest", "sendMessage")
 	core.RegisterCallback(self, "addPlayerQuestObjective", "sendMessage")
	core.RegisterCallback(self, "endPlayerQuest", "sendMessage")
	core.RegisterCallback(self, "playerQuestObjectiveProgress", "sendMessage")
	core.RegisterCallback(self, "playerQuestComplete", "sendMessage")
	core.RegisterCallback(self, "playerQuestFailed", "sendMessage")
	core.RegisterCallback(self, "playerQuestLost", "sendMessage")
	core.RegisterCallback(self, "endSync", "sendMessage")
	core.RegisterCallback(self, "initParty", "PARTY_MEMBERS_CHANGED")
end

function module:OnDisable()
	core.UnregisterAllCallbacks(self)
end

function module:PARTY_MEMBERS_CHANGED()
	core.numPartyMembers = GetNumPartyMembers() or 0
	local timeStamp = GetTime()
	local partySync = 0
	if core.numPartyMembers > 0 then
		for i=1, core.numPartyMembers do
			local partyName = UnitName('party'..i)
			if not core.party[partyName] then
				core.party[partyName] = {}
				core.party[partyName].status = 0
			end
			if not core.party[partyName].timestamp then
				local _, partyClass = UnitClass('party'..i)
				local classColor = RAID_CLASS_COLORS[partyClass]
				core.party[partyName].color = format("%02x%02x%02x", classColor.r*255, classColor.g*255, classColor.b*255)
				module:SendCommMessage(addonPrefix, module:Serialize("v", version, firstCompatibleVersion), "WHISPER", partyName)
			end
			core.party[partyName].timestamp = timeStamp
			partySync = partySync + core.party[partyName].status
		end
	end

	for partyName in pairs(core.party) do
		if core.party[partyName].timestamp ~= timeStamp then
			core.cbh:Fire("initTrackedQuest")
			for uid in pairs(core.quests) do
				if core.quests[uid].party[partyName] then
					core:questLost(partyName, false, uid)
				end
			end
			delTable(core.party[partyName])
			core.party[partyName] = nil
		end
	end
	if partySync == 0 then
		syncParty = false
	else
		syncParty = true
	end
end
-------------
--- INPUT ---
-------------
local comm = {
	v = function(sender, senderVersion, senderFirstCompatibleVersion)
		if senderVersion <= lastIncompatibleVersion then
			DEFAULT_CHAT_FRAME:AddMessage(format(L["The tomQuest2 release (%d) used by %s is too old to synchronise with your release (%d) (min %d)"], senderVersion, sender, version, firstCompatibleVersion))
			return
		elseif version < senderFirstCompatibleVersion then
			DEFAULT_CHAT_FRAME:AddMessage(format(L["Your tomQuest2 release (%d) is too old to synchorise with release (%d) of %s (min %d)"], version, senderVersion, sender, firstCompatibleVersion))
			return
		elseif version < senderVersion then
			DEFAULT_CHAT_FRAME:AddMessage(format(L["New release available (%d) ask %s where to find it"], senderVersion, sender))
		end
		core.party[sender] = core.party[sender] or {}
		core.party[sender].status = 1
		syncParty = true
		for uid in pairs(core.quests) do
			if core.quests[uid].party[core.playerName] then
				module:SendCommMessage(addonPrefix, module:Serialize("q", uid, core.quests[uid].title, core.quests[uid].shorttag, core.quests[uid].level, core.quests[uid].numobjectives, core.quests[uid].party[core.playerName].complete or 0), "WHISPER", sender, "BULK")
				if core.questsObjectives[uid] then
					for objective in pairs(core.questsObjectives[uid]) do
						if core.questsObjectives[uid][objective].party[core.playerName] then
							module:SendCommMessage(addonPrefix, module:Serialize("o", uid, objective, core.questsObjectives[uid][objective].desc, core.questsObjectives[uid][objective].qtype, core.questsObjectives[uid][objective].party[core.playerName].got, core.questsObjectives[uid][objective].need, core.questsObjectives[uid][objective].party[core.playerName].ngot, core.questsObjectives[uid][objective].nneed), "WHISPER", sender, "BULK")
						end
					end
				end
			end
		end
		module:SendCommMessage(addonPrefix, module:Serialize("d"), "WHISPER", sender, "BULK")
	end,
	q = function(sender, uid, title, shorttag, level, numobjectives, complete)
		if not core.party[sender] then return end
		if not core.quests[uid] and not core.questsZones[L["PARTYQUESTSZONE"]] then
			core:addZone(L["PARTYQUESTSZONE"])
		end
		core:addQuest(sender, false, uid, title, shorttag, level, numobjectives, complete, L["PARTYQUESTSZONE"], nil)
	end,
	o = function(sender, uid, objective, desc, qtype, got, need, ngot, nneed)
		if not core.party[sender] or not core.quests[uid] or not core.quests[uid].party[sender] or not core.questsObjectives[uid] then return end
		core:addQuestObjective(sender, false, uid, objective, desc, qtype, got, need, ngot, nneed)
	end,
	r = function(sender, uid)
		if not core.party[sender] or not core.quests[uid] or not core.quests[uid].party[sender] then return end
		core.cbh:Fire("updateQuestsTooltip")
	end,
	m = function(sender, uid)
		if not core.party[sender] or not core.quests[uid] or not core.quests[uid].party[sender] or core.quests[uid].objective then return end
		module:SendCommMessage(addonPrefix, module:Serialize("n", uid, core.quests[uid].objective), "WHISPER", sender, "BULK")
	end,
	n = function(sender, uid, objective)
		if not core.party[sender] or not core.quests[uid] or core.quests[uid].objective then return end
		core.quests[uid].objective = objective
		core.cbh:Fire("questObjectiveAdded", uid)
	end,
	s = function(sender, uid)
		if not core.party[sender] or not core.quests[uid] or not core.quests[uid].party[sender] or core.quests[uid].description then return end
		module:SendCommMessage(addonPrefix, module:Serialize("t", uid, core.quests[uid].description), "WHISPER", sender, "BULK")
	end,
	t = function(sender, uid, description)
		if not core.party[sender] or not core.quests[uid] or core.quests[uid].description then return end
		core.quests[uid].description = description
		core.cbh:Fire("questDescriptionAdded", uid)
	end,
	p = function(sender, uid, objective, got, ngot)
		if not core.party[sender] or not core.quests[uid] then return end
		core:questObjectiveProgress(sender, false, uid, objective, got, ngot)
	end,
	c = function(sender, uid)
		if not core.party[sender] or not core.quests[uid] or not core.quests[uid].party[sender] then return end
		core:questComplete(sender, false, uid)
	end,
	f = function(sender, uid)
		if not core.party[sender] or not core.quests[uid] or not core.quests[uid].party[sender] then return end
		core:questFailed(sender, false, uid)
	end,
	l = function(sender, uid)
		if not core.party[sender] or not core.quests[uid] or not core.quests[uid].party[sender] then return end
		core:questLost(sender, false, uid)
	end,
	d = function(sender)
		if not core.party[sender] then return end
		core.cbh:Fire("updateQuestsTracker")
		core.cbh:Fire("updateQuestsTooltip")
	end,
}

function module:commHandler(prefix, message, distribution, sender)
	if sender ~= core.playerName then
		module:onCommand(sender, module:Deserialize(message))
	end
end

function module:onCommand(sender, success, commHandler, ...)
	if not success then
		return
	elseif comm[commHandler] then
		comm[commHandler](sender, ...)
	end
end

function module:checkQuestObjective(arg)
	if type(arg) ~= "table" or not arg[1] or not arg[2] or not core.quests[arg[1]] or not core.party[arg[2]] then return end
	if not core.quests[arg[1]].objective then
		module:SendCommMessage(addonPrefix, module:Serialize("m", arg[1]), "WHISPER", arg[2], "BULK")
	end
	if not core.quests[arg[1]].description then
		module:SendCommMessage(addonPrefix, module:Serialize("s", arg[1]), "WHISPER", arg[2], "BULK")
	end
end
--------------
--- OUTPUT ---
--------------
function module:sendMessage(event, firstRun, ...)
	if core.numPartyMembers > 0 and syncParty and not firstRun then
		module:SendCommMessage(addonPrefix, module:Serialize(...), "PARTY", nil, "BULK")
	end
end